import rapidjson
import pytest

from nephthys import RequestLogRecord, LogRecord
from nephthys.filters.requests import HeaderFilter, BodyTypeFilter, JsonBodyFilter, QueryStringFilter
from nephthys.filters.requests import (
    RequestType,
    QS_FILTERED,
    HEADER_FILTERED,
    BODY_NOT_LOGGABLE,
    JSON_BODY_FILTERED,
)


def rec_generator():
    return LogRecord(extra_tags=["my_tag"])


def req_rec_generator(
    request_headers=None, response_headers=None, request_body=None, response_body=None, qs=None
):
    req_rec = RequestLogRecord()

    for name, value in request_headers or []:
        req_rec.add_request_header(name, value)
    for name, value in response_headers or []:
        req_rec.add_response_header(name, value)

    for name, value in qs or []:
        req_rec.add_request_querystring(name, value)

    req_rec.request_body = request_body
    req_rec.response_body = response_body

    return req_rec


@pytest.mark.parametrize(
    "filters,in_record,out_record",
    [
        (
            ["QSFiltered", "QSfiltered2"],
            req_rec_generator(
                qs=[
                    ("QSFiltered", "value1"),
                    ("QSfiltered", "value2"),
                    ("QSNotFiltered", "value3"),
                    ("QSfiltered2", "value4"),
                ],
            ),
            req_rec_generator(
                qs=[
                    ("QSFiltered", QS_FILTERED),
                    ("QSfiltered", "value2"),
                    ("QSNotFiltered", "value3"),
                    ("QSfiltered2", QS_FILTERED),
                ],
            ),
        ),
        (
            [],
            req_rec_generator(
                qs=[
                    ("QSFiltered", "value1"),
                    ("QSfiltered", "value2"),
                    ("QSNotFiltered", "value3"),
                    ("QSfiltered2", "value4"),
                ],
            ),
            req_rec_generator(
                qs=[
                    ("QSFiltered", "value1"),
                    ("QSfiltered", "value2"),
                    ("QSNotFiltered", "value3"),
                    ("QSfiltered2", "value4"),
                ],
            ),
        ),
        (["QSFiltered"], rec_generator(), rec_generator()),
    ],
)
def test_qs_filter(filters, in_record, out_record):
    qs_filter = QueryStringFilter(filters)

    qs_filter.filter(in_record)

    assert in_record.asdict() == out_record.asdict()


@pytest.mark.parametrize(
    "filters,req_type,in_record,out_record",
    [
        (
            ["X-Filter-me"],
            RequestType.REQUEST,
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
                response_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
            ),
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", HEADER_FILTERED),
                    ("X-Not-Filter-Me", "value2"),
                ],
                response_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
            ),
        ),
        (
            ["X-Filter-Me"],
            RequestType.RESPONSE,
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
                response_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
            ),
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
                response_headers=[
                    ("X-Filter-Me", HEADER_FILTERED),
                    ("X-Not-Filter-Me", "value2"),
                ],
            ),
        ),
        (
            ["X-Filter-Me"],
            RequestType.ALL,
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
                response_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
            ),
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", HEADER_FILTERED),
                    ("X-Not-Filter-Me", "value2"),
                ],
                response_headers=[
                    ("X-Filter-Me", HEADER_FILTERED),
                    ("X-Not-Filter-Me", "value2"),
                ],
            ),
        ),
        (
            ["X-Filter-Me"],
            RequestType.ALL,
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ]
            ),
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", HEADER_FILTERED),
                    ("X-Not-Filter-Me", "value2"),
                ]
            ),
        ),
        (
            [],
            RequestType.ALL,
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
                response_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
            ),
            req_rec_generator(
                request_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
                response_headers=[
                    ("X-Filter-Me", "value1"),
                    ("X-Not-Filter-Me", "value2"),
                ],
            ),
        ),
        (["X-Filter-Me"], RequestType.ALL, rec_generator(), rec_generator()),
    ],
)
def test_header_filter(filters, req_type, in_record, out_record):
    head_filter = HeaderFilter(filters, req_type)

    head_filter.filter(in_record)

    assert in_record.asdict() == out_record.asdict()


@pytest.mark.parametrize(
    "filters,req_type,in_record,out_record",
    [
        (
            None,
            RequestType.REQUEST,
            req_rec_generator(
                request_headers=[("Content-Type", "multipart/form-data")],
                response_headers=[("Content-Type", "image/jpeg")],
                request_body="Filter-Me",
                response_body="Not-Filter-Me",
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "multipart/form-data")],
                response_headers=[("Content-Type", "image/jpeg")],
                request_body=BODY_NOT_LOGGABLE.format("multipart/form-data"),
                response_body="Not-Filter-Me",
            ),
        ),
        (
            None,
            RequestType.RESPONSE,
            req_rec_generator(
                request_headers=[("Content-Type", "image/jpeg")],
                response_headers=[("Content-Type", "multipart/form-data")],
                request_body="Not-Filter-Me",
                response_body="Filter-Me",
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "image/jpeg")],
                response_headers=[("Content-Type", "multipart/form-data")],
                request_body="Not-Filter-Me",
                response_body=BODY_NOT_LOGGABLE.format("multipart/form-data"),
            ),
        ),
        (
            None,
            RequestType.ALL,
            req_rec_generator(
                request_headers=[("Content-Type", "multipart/form-data")],
                response_headers=[("Content-Type", "image/jpeg")],
                request_body="Filter-Me",
                response_body="Filter-Me",
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "multipart/form-data")],
                response_headers=[("Content-Type", "image/jpeg")],
                request_body=BODY_NOT_LOGGABLE.format("multipart/form-data"),
                response_body=BODY_NOT_LOGGABLE.format("image/jpeg"),
            ),
        ),
        (
            ["application/json"],
            RequestType.ALL,
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "text/plain")],
                request_body="Not-Filter-Me",
                response_body="Filter-Me",
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "text/plain")],
                request_body="Not-Filter-Me",
                response_body=BODY_NOT_LOGGABLE.format("text/plain"),
            ),
        ),
        (
            ["application/json"],
            RequestType.ALL,
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                request_body="Not-Filter-Me",
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                request_body="Not-Filter-Me",
            ),
        ),
        (["application/json"], RequestType.ALL, rec_generator(), rec_generator()),
    ],
)
def test_body_type_filter(filters, req_type, in_record, out_record):
    head_filter = BodyTypeFilter(filters, req_type)

    head_filter.filter(in_record)

    assert in_record.asdict() == out_record.asdict()


@pytest.mark.parametrize(
    "filters,req_type,in_record,out_record",
    [
        (
            {"key": True},
            RequestType.REQUEST,
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "application/json")],
                request_body=rapidjson.dumps(
                    {"key": "Filter-Me", "key2": "Not-Filter-Me"}
                ),
                response_body=rapidjson.dumps({"key": "Not-filter-Me"}),
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "application/json")],
                request_body=rapidjson.dumps(
                    {"key": JSON_BODY_FILTERED, "key2": "Not-Filter-Me"}
                ),
                response_body=rapidjson.dumps({"key": "Not-filter-Me"}),
            ),
        ),
        (
            {"key": True},
            RequestType.RESPONSE,
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "application/json")],
                request_body=rapidjson.dumps({"key": "Not-filter-Me"}),
                response_body=rapidjson.dumps(
                    {"key": "Filter-Me", "key2": "Not-Filter-Me"}
                ),
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "application/json")],
                request_body=rapidjson.dumps({"key": "Not-filter-Me"}),
                response_body=rapidjson.dumps(
                    {"key": JSON_BODY_FILTERED, "key2": "Not-Filter-Me"}
                ),
            ),
        ),
        (
            {"key": {"key": True}},
            RequestType.ALL,
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "application/json")],
                request_body=rapidjson.dumps({"key": {"key": "Filter-Me"}}),
                response_body=rapidjson.dumps(
                    {"key": {"key": "Filter-Me"}, "key2": "Not-Filter-Me"}
                ),
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "application/json")],
                request_body=rapidjson.dumps({"key": {"key": JSON_BODY_FILTERED}}),
                response_body=rapidjson.dumps(
                    {"key": {"key": JSON_BODY_FILTERED}, "key2": "Not-Filter-Me"}
                ),
            ),
        ),
        (
            {"key": {"key": True}},
            RequestType.ALL,
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "application/json")],
                request_body=rapidjson.dumps({"key": "Not-Filter-Me"}),
                response_body=rapidjson.dumps({"key2": "Not-Filter-Me"}),
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "application/json")],
                response_headers=[("Content-Type", "application/json")],
                request_body=rapidjson.dumps({"key": "Not-Filter-Me"}),
                response_body=rapidjson.dumps({"key2": "Not-Filter-Me"}),
            ),
        ),
        (
            {"key": {"key": True}},
            RequestType.ALL,
            req_rec_generator(
                request_headers=[("Content-Type", "text/plain")],
                response_headers=[("Content-Type", "application/json")],
                request_body="Not-Filter-Me",
                response_body=rapidjson.dumps({"key2": "Not-Filter-Me"}),
            ),
            req_rec_generator(
                request_headers=[("Content-Type", "text/plain")],
                response_headers=[("Content-Type", "application/json")],
                request_body="Not-Filter-Me",
                response_body=rapidjson.dumps({"key2": "Not-Filter-Me"}),
            ),
        ),
        ({"key": {"key": True}}, RequestType.ALL, rec_generator(), rec_generator()),
    ],
)
def test_json_body_filter(filters, req_type, in_record, out_record):
    head_filter = JsonBodyFilter(filters, req_type)

    head_filter.filter(in_record)

    assert in_record.asdict() == out_record.asdict()



import itertools
from logging import LoggerAdapter
from webob.multidict import MultiDict
from urllib.parse import urlparse


class BaseLoggerAdapter(LoggerAdapter):
    def __init__(self, logger, extra_tags=None, *args, **kwargs):
        super().__init__(logger=logger, extra=None, *args, **kwargs)

        self._extra_tags = extra_tags or []
        self._extra_tags.append(logger.name)

    def _process(self, msg):
        if isinstance(msg, str):
            msg = Log(LogRecord(message=msg))

        if isinstance(msg, Log):
            log_rec = msg.log_record
            log_rec.add_tags(self._extra_tags)

        return msg

    def process(self, msg, kwargs):

        proc_msg = self._process(msg)

        if isinstance(proc_msg, Log):
            return proc_msg.log_record.asdict(), kwargs

        return proc_msg, kwargs


def apply_filters(log_record, filters):
    for f in filters:
        if hasattr(f, "filter"):
            f.filter(log_record)
        else:
            f(log_record)


class FilterLoggerAdapter(BaseLoggerAdapter):
    def __init__(self, logger, filters=None, *args, **kwargs):
        super().__init__(logger=logger, *args, **kwargs)

        self._filters = filters or []

    def _process(self, msg):
        msg = super()._process(msg)

        if isinstance(msg, Log):
            log_rec = msg.log_record
            filters = self._filters

            if isinstance(msg, FilterableLog):
                filters = itertools.chain(filters, msg.filters)

            apply_filters(log_rec, filters)

        return msg

    def log(self, level, msg, *args, **kwargs):
        if isinstance(msg, FilterableLog) and msg.drop:
            return

        super().log(level, msg, *args, **kwargs)


def join_multidict(multi_dict):
    list_dict = multi_dict.dict_of_lists()
    return {key: ",".join(value) for key, value in list_dict.items()}


def add_to_multidict(multi_dict, name, value):
    if isinstance(value, list):
        for val in value:
            multi_dict.add(name, str(val))
    else:
        multi_dict.add(name, str(value))


class Log:
    def __init__(self, log_record):
        self._log_record = log_record

    @property
    def log_record(self):
        return self._log_record


class FilterableLog(Log):
    def __init__(self, log_record):
        super().__init__(log_record)
        self.drop = False
        self._filters = []

    @property
    def filters(self):
        return self._filters

    def add_filters(self, record_filters):
        if isinstance(record_filters, list):
            self._filters.extend(record_filters)
        else:
            self._filters.append(record_filters)


class LogRecord:
    def __init__(self, message="", extra_tags=None, *args, **kwargs):
        self._extra_tags = extra_tags or []
        self._message = message

    def asdict(self):
        msg_dict = {"extra_tags": self._extra_tags, "message": self._message}

        return msg_dict

    def add_tags(self, tags):
        if isinstance(tags, list):
            self._extra_tags.extend(tags)
        else:
            self._extra_tags.append(tags)


class RequestLogRecord(LogRecord):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._req_start = None
        self._req_end = None
        self._req_time = None
        self._method = None
        self._url = None
        self._path = None
        self._host = None
        self._route = None
        self._status_code = None
        self._user = None
        self._user_uuid = None
        self._req_query = MultiDict()
        self._req_headers = MultiDict()
        self._req_body = None
        self._res_headers = MultiDict()
        self._res_body = None
        self._route_match = {}

    def asdict(self):
        base_dict = super().asdict()

        req_dict = {
            "request": {
                "start": self._req_start,
                "end": self._req_end,
                "time": self._req_time,
                "method": self._method,
                "header": join_multidict(self._req_headers),
                "query": join_multidict(self._req_query),
                "url": self._url,
                "host": self._host,
                "path": self._path,
                "route": self._route,
                "route_match": self._route_match,
                "user": self._user,
                "user_uuid": self._user_uuid,
                "body": self._req_body,
            },
            "response": {
                "status_code": self._status_code,
                "header": join_multidict(self._res_headers),
                "body": self._res_body,
            },
        }

        return {**base_dict, **req_dict}

    def add_request_querystring(self, name, value):
        add_to_multidict(self._req_query, name, value)

    def add_request_header(self, name, value):
        name = name.title()
        add_to_multidict(self._req_headers, name, value)

    def add_response_header(self, name, value):
        name = name.title()
        add_to_multidict(self._res_headers, name, value)

    def add_route_match(self, name, value):
        self._route_match[name] = value

    def _set_request_start(self, value):
        self._req_start = value

        if self._req_end:
            self._req_time = (self._req_end - self._req_start) * 1000

    def _set_request_end(self, value):
        self._req_end = value

        if self._req_start:
            self._req_time = (self._req_end - self._req_start) * 1000

    def _set_request_body(self, body):
        self._req_body = body

    def _set_response_body(self, body):
        self._res_body = body

    def _set_method(self, value):
        self._method = value.upper()

    def _set_url(self, value):
        parsed_url = urlparse(value)

        self._url = value
        self._path = parsed_url.path
        self._host = parsed_url.netloc

    def _set_route(self, value):
        self._route = value

    def _set_status_code(self, value):
        _val = int(value)
        if _val > 599 or _val < 100:
            raise ValueError

        self._status_code = int(value)

    def _set_user(self, value):
        self._user = value

    def _set_user_uuid(self, value):
        self._user_uuid = value

    request_start = property(None, _set_request_start)
    request_end = property(None, _set_request_end)
    request_body = property(None, _set_request_body)
    response_body = property(None, _set_response_body)
    method = property(None, _set_method)
    url = property(None, _set_url)
    route = property(None, _set_route)
    status_code = property(None, _set_status_code)
    user = property(None, _set_user)
    user_uuid = property(None, _set_user_uuid)



import logging
import rapidjson
import re


class JSONRenderer:
    def __init__(self, sort_keys=False, indent=None):
        self._sort_keys = sort_keys
        self._indent = indent

    def __call__(self, content):
        return rapidjson.dumps(
            content,
            default=self.default,
            sort_keys=self._sort_keys,
            indent=self._indent,
        )

    def default(self, content):
        return str(content)


class JSONFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        self.render_exc = kwargs.pop("render_exc", True)

        sort_keys = kwargs.pop("json_sort_keys", False)
        indent = kwargs.pop("json_indent", None)

        super().__init__(*args, **kwargs)

        self._renderer = JSONRenderer(indent=indent, sort_keys=sort_keys)
        self._required_fields = self._parse()

    def _parse(self):
        """
        Parses format string looking for substitutions
        This method is responsible for returning a list of fields (as strings)
        to include in all log messages.
        """
        standard_formatters = re.compile(r"\((.+?)\)", re.IGNORECASE)
        return standard_formatters.findall(self._fmt)

    def _add_fields(self, log_record, record, message_dict):
        """
        Override this method to implement custom logic for adding fields.
        """
        for field in self._required_fields:
            log_record[field] = record.__dict__.get(field)
        log_record.update(message_dict)

    def format(self, record):
        """Formats a log record and serializes to json"""
        message_dict = {}
        if isinstance(record.msg, dict):
            message_dict = record.msg
        else:
            record.message = record.getMessage()

        if self.render_exc:
            # Display formatted exception, but allow overriding it in the
            # user-supplied dict.
            if record.exc_info and not message_dict.get("exc_info"):
                message_dict["exc_info"] = self.formatException(record.exc_info)
                message_dict["exc"] = record.exc_info[0].__name__
            if not message_dict.get("exc_info") and record.exc_text:
                message_dict["exc_info"] = record.exc_text
            # Display formatted record of stack frames
            # default format is a string returned from :func:`traceback.print_stack`
            if record.stack_info and not message_dict.get("stack_info"):
                message_dict["stack_info"] = self.formatStack(record.stack_info)

        log_record = {}

        self._add_fields(log_record, record, message_dict)

        return self._renderer(log_record)



import logging
import pytest

from nephthys import BaseLoggerAdapter, Log, LogRecord


@pytest.fixture
def logger():
    return logging.getLogger("test_logger")


def log_generator(message=None, extra_tags=None):
    return Log(LogRecord(message=message, extra_tags=extra_tags))


@pytest.mark.parametrize(
    "tags,message,output",
    [
        ([], "test", {"extra_tags": ["test_logger"], "message": "test"}),
        (
            ["tag1", "tag2"],
            "",
            {"extra_tags": ["tag1", "tag2", "test_logger"], "message": ""},
        ),
        (
            ["tag1", "tag2"],
            "test",
            {"extra_tags": ["tag1", "tag2", "test_logger"], "message": "test"},
        ),
        (
            ["tag1", "tag2"],
            log_generator(message="test", extra_tags=["log_rec_tag"]),
            {
                "extra_tags": ["log_rec_tag", "tag1", "tag2", "test_logger"],
                "message": "test",
            },
        ),
    ],
)
def test_process(logger, tags, message, output):
    base_logger = BaseLoggerAdapter(logger, tags)
    log_dict, kwargs = base_logger.process(message, {})

    assert log_dict == output



import re

from .filter import IFilter
from .. import LogRecord

FILTER_STRING = "<filtered>"


class MessageBlacklist(IFilter):
    def __init__(self, blacklist=None):
        self._blacklist = blacklist or []
        self._blacklist_string = "|".join(self._blacklist)

    def filter(self, log_record):
        if not isinstance(log_record, LogRecord):
            return

        log_record._message = re.sub(
            self._blacklist_string, FILTER_STRING, log_record._message
        )



import rapidjson
from enum import Enum

from .filter import IFilter
from .. import RequestLogRecord


QS_FILTERED = "<filtered>"
HEADER_FILTERED = "<filtered>"
LOGGABLE_TYPES = ["application/json", "text/plain", "text/html"]
BODY_NOT_LOGGABLE = "<body not loggable Content-Type {}>"
JSON_BODY_FILTERED = "<filtered>"


def filter_json_body(s, r):
    for k, v in s.items():
        if k in r:
            if isinstance(v, dict):
                filter_json_body(v, r[k])
            else:
                r[k] = JSON_BODY_FILTERED


def find_content_type(headers):
    return ",".join(headers.getall("Content-Type"))


class RequestType(Enum):
    ALL = 1
    REQUEST = 2
    RESPONSE = 3


class HeaderFilter(IFilter):
    def __init__(self, headers=None, req_type=RequestType.ALL):
        self._headers = headers or []
        self._req_type = req_type

    def _filter_headers(self, headers):
        for hf in self._headers:
            hft = hf.title()
            if hft in headers:
                headers[hft] = HEADER_FILTERED

    def filter(self, log_record):
        if not isinstance(log_record, RequestLogRecord):
            return

        if self._req_type == RequestType.REQUEST or self._req_type == RequestType.ALL:
            self._filter_headers(log_record._req_headers)

        if self._req_type == RequestType.RESPONSE or self._req_type == RequestType.ALL:
            self._filter_headers(log_record._res_headers)


class QueryStringFilter(IFilter):
    def __init__(self, keys=None):
        self._keys = keys or []

    def _filter_keys(self, keys):
        for kf in self._keys:
            if kf in keys:
                keys[kf] = QS_FILTERED

    def filter(self, log_record):
        if not isinstance(log_record, RequestLogRecord):
            return

        self._filter_keys(log_record._req_query)


class BodyTypeFilter(IFilter):
    def __init__(self, allowed_types=None, req_type=RequestType.ALL):
        self._allowed_types = LOGGABLE_TYPES if allowed_types is None else allowed_types
        self._req_type = req_type

    def _filter_body(self, body, content_type):
        if content_type and any(valid_type in content_type for valid_type in self._allowed_types):
            return body

        return BODY_NOT_LOGGABLE.format(content_type)

    def filter(self, log_record):
        if not isinstance(log_record, RequestLogRecord):
            return

        if log_record._req_body and (
            self._req_type == RequestType.REQUEST or self._req_type == RequestType.ALL
        ):
            content_type = find_content_type(log_record._req_headers)
            log_record._req_body = self._filter_body(
                log_record._req_body, content_type
            )

        if log_record._res_body and (
            self._req_type == RequestType.RESPONSE or self._req_type == RequestType.ALL
        ):
            content_type = find_content_type(log_record._res_headers)
            log_record._res_body = self._filter_body(
                log_record._res_body, content_type
            )


class JsonBodyFilter(IFilter):
    def __init__(self, body_schema, req_type=RequestType.ALL):
        self._body_schema = body_schema or {}
        self._req_type = req_type

    def filter(self, log_record):
        if not isinstance(log_record, RequestLogRecord):
            return

        if log_record._req_body and (
            self._req_type == RequestType.REQUEST or self._req_type == RequestType.ALL
        ):
            if "application/json" in find_content_type(log_record._req_headers):
                req_body = rapidjson.loads(log_record._req_body)
                filter_json_body(self._body_schema, req_body)
                log_record._req_body = rapidjson.dumps(req_body)

        if log_record._res_body and (
            self._req_type == RequestType.RESPONSE or self._req_type == RequestType.ALL
        ):
            if "application/json" in find_content_type(log_record._res_headers):
                res_body = rapidjson.loads(log_record._res_body)
                filter_json_body(self._body_schema, res_body)
                log_record._res_body = rapidjson.dumps(res_body)



import logging
from datetime import datetime
from urllib.parse import parse_qs, urlparse
from requests.sessions import Session as RequestsSession

from nephthys import FilterLoggerAdapter, RequestLogRecord, Log
from nephthys.filters.requests import BodyTypeFilter

logger = logging.getLogger("requests_out")

DEFAULT_ALLOWED_TYPES = [
    "application/json",
    "text/plain",
    "text/html",
    "application/x-www-form-urlencoded",
]


def catch_logger_exception(function):
    def wrapper(*args, **kwargs):
        try:
            function(*args, **kwargs)
        except Exception as exc:
            logger.exception("Failed to log")

    return wrapper


def decorate_log_request(log_record, request):
    log_record.method = request.method
    log_record.url = request.url

    if request.headers:
        for name, value in request.headers.items():
            log_record.add_request_header(name, value)

    querystring = parse_qs(urlparse(request.url).query)
    if querystring:
        for name, value in querystring.items():
            log_record.add_request_querystring(name, value)
    if request.body:
        try:
            if isinstance(request.body, str):
                log_record.request_body = request.body
            else:
                log_record.request_body = request.body.decode("UTF-8", errors="strict")
        except UnicodeDecodeError:
            log_record.request_body = "<RAW Data>"


def decorate_log_response(log_record, response):
    log_record.status_code = response.status_code

    if response.headers:
        for name, value in response.headers.items():
            log_record.add_response_header(name, value)

    if response.content:
        encoding = response.encoding or "utf-8"

        try:
            log_record.response_body = response.content.decode(
                encoding, errors="strict"
            )
        except (TypeError, LookupError, UnicodeDecodeError):
            log_record.response_body = "<RAW Data>"


class NephthysMixin:
    """
    Provides an easy way to add a nephthys logger a class.
    The logger is configured with a default nephthys.filters.requests.BodyTypeFilter,
    instantiated with nephthys.clients.requests_session.DEFAULT_ALLOWED_TYPES as its
    allowed_types.
    """

    _logger = None

    def __init__(self, log_tag=None, log_filters=None, *args, **kwargs):
        """
        :param log_tag: The tag that will identify logs from this Session
        :type log_tag: string
        :param log_filters: List of additional IRequestFilter
        :type log_filters: list
        """
        _log_filters = [BodyTypeFilter(allowed_types=DEFAULT_ALLOWED_TYPES)]

        if isinstance(log_filters, list):
            _log_filters.extend(log_filters)

        self._logger = FilterLoggerAdapter(
            logger=logger, filters=_log_filters, extra_tags=[log_tag]
        )
        super().__init__(*args, **kwargs)

    @catch_logger_exception
    def _send_log_record(
        self, start_time, end_time, request=None, response=None, exception=False
    ):
        """
        Builds a LogRecord and logs it with self._logger.
        Integrate this method in your class to enable Nephthys logging into it.
        :param start_time: timestamp of when the operation logged started
        :param end_time: timestamp of when the operation logged ended
        :type request: requests.models.Request
        :type response: requests.models.Response
        :param exception: whether the log is an exception or not
        """
        log_rec = RequestLogRecord()
        log_rec.request_start = start_time
        log_rec.request_end = end_time

        decorate_log_request(log_rec, request)

        if response is not None:
            decorate_log_response(log_rec, response)

        if exception:
            self._logger.exception(Log(log_rec))
        else:
            self._logger.info(Log(log_rec))

    def send(self, request, **kwargs):
        start_time = datetime.utcnow().timestamp()

        try:
            response = super().send(request, **kwargs)
        except Exception as exc:
            self._send_log_record(
                start_time=start_time,
                end_time=datetime.utcnow().timestamp(),
                request=request,
                exception=True,
            )
            raise

        self._send_log_record(
            start_time=start_time,
            end_time=datetime.utcnow().timestamp(),
            request=request,
            response=response,
        )

        return response


class Session(NephthysMixin, RequestsSession):
    """
    Provides a requests.session.Session with Nephthys Logging.
    """
    pass



import json
import sys
from logging import LogRecord

import pytest

from nephthys.formatters.json import JSONRenderer, JSONFormatter


@pytest.fixture
def msg():
    return {"test": "test"}


def serializable_obj():
    return


def test_renderer():
    renderer = JSONRenderer()
    assert renderer({"key": 2 + 1j}) == '{"key":"(2+1j)"}'
    assert renderer({"test": "test"}) == '{"test":"test"}'
    assert renderer({"b": "test", "a": "test"}) != '{"a":"test","b": "test"}'


def test_renderer_sort():
    renderer = JSONRenderer(sort_keys=True)
    assert renderer({"b": "test", "a": "test"}) == '{"a":"test","b":"test"}'


def test_renderer_indent():
    renderer = JSONRenderer(indent=1)
    assert renderer({"test": "test"}) == '{\n "test": "test"\n}'


def test_renderer_default():
    renderer = JSONRenderer()
    assert renderer("test") == '"test"'


def test_defaults():
    formatter = JSONFormatter(fmt="$(message)s")

    assert isinstance(formatter._renderer, JSONRenderer)
    assert formatter._required_fields == ["message"]


def test_not_a_dict():
    formatter = JSONFormatter()
    log = LogRecord("test", 20, "/app/module", 1, "My message", [], None)
    out = formatter.format(log)

    assert out == '{"message":"My message"}'


def test_default(msg):
    formatter = JSONFormatter()
    log = LogRecord("test", 20, "/app/module", 1, msg, [], None)
    out = formatter.format(log)

    assert out == '{"message":null,"test":"test"}'


def test_required_parameters(msg):
    formatter = JSONFormatter(fmt="$(name)s $(levelno)s $(pathname)s $(lineno)s")
    log = LogRecord("test", 20, "/app/module", 1, msg, [], None)
    out = formatter.format(log)

    assert (
        out
        == '{"name":"test","levelno":20,"pathname":"/app/module","lineno":1,"test":"test"}'
    )


def test_exception(msg):
    formatter = JSONFormatter()

    try:
        raise Exception
    except Exception:
        exc_info = sys.exc_info()

        log = LogRecord(
            "test", 20, "/app/module", 1, msg, [], exc_info, sinfo="Stack Informations"
        )
        out = formatter.format(log)
        out_dict = json.loads(out)

        assert out_dict["exc_info"] == formatter.formatException(exc_info)
        assert out_dict["exc"] == "Exception"
        assert out_dict["stack_info"] == "Stack Informations"
        assert out_dict["test"] == "test"



from abc import ABC, abstractmethod


class IFilter(ABC):
    @abstractmethod
    def filter(self, log_record):
        """
        Applies configured filters to log_record
        :param log_record: The record that needs to be filtered
        :type log_record: LogRecord
        """
        pass  # pragma: nocover



import io

from setuptools import setup, find_packages


def make_long_description():
    with io.open("README.md", encoding="utf-8") as fp:
        long_description = fp.read()
    return long_description


setup(
    name="nephthys",
    description="Advanced Python Logger",
    long_description=make_long_description(),
    long_description_content_type="text/markdown",
    version="1.0.1",
    author="Fabio Todaro",
    license="MIT",
    author_email="ft@ovalmoney.com",
    url="https://github.com/OvalMoney/Nephthys",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=["tests", "requirements"]),
    install_requires=["webob"],
    extras_require={"JSON": ["python-rapidjson"], "requests": ["requests"]},
)



import logging
import pytest

from nephthys.filters.message import MessageBlacklist
from nephthys import FilterLoggerAdapter, FilterableLog, Log, LogRecord


@pytest.fixture
def logger():
    return logging.getLogger("test_logger")


def log_generator(message=None, extra_tags=None):
    return Log(LogRecord(message=message, extra_tags=extra_tags))


def filt_log_generator(record_filter=None, message=None, extra_tags=None):
    filt_log = FilterableLog(LogRecord(message=message, extra_tags=extra_tags))
    if record_filter:
        filt_log.add_filters([record_filter])
    return filt_log


@pytest.mark.parametrize(
    "filters,message,output",
    [
        ([], "test", {"extra_tags": ["test_logger"], "message": "test"}),
        ([], "", {"extra_tags": ["test_logger"], "message": ""}),
        ([], "test", {"extra_tags": ["test_logger"], "message": "test"}),
        (
            [MessageBlacklist(["blacklisted"])],
            "blacklisted test blacklisted",
            {"extra_tags": ["test_logger"], "message": "<filtered> test <filtered>"},
        ),
        (
            [],
            log_generator(message="test", extra_tags=["log_rec_tag"]),
            {"extra_tags": ["log_rec_tag", "test_logger"], "message": "test"},
        ),
        (
            [MessageBlacklist(["blacklisted"])],
            log_generator(message="blacklisted test blacklisted"),
            {"extra_tags": ["test_logger"], "message": "<filtered> test <filtered>"},
        ),
        (
            [MessageBlacklist(["blacklisted_1"])],
            filt_log_generator(
                record_filter=MessageBlacklist(["blacklisted_2"]),
                message="blacklisted_1 test blacklisted_2",
            ),
            {"extra_tags": ["test_logger"], "message": "<filtered> test <filtered>"},
        ),
    ],
)
def test_process(logger, filters, message, output):
    base_logger = FilterLoggerAdapter(logger, filters=filters)
    log_dict, kwargs = base_logger.process(message, {})

    assert log_dict == output



import logging
from datetime import datetime

import requests_mock
from freezegun import freeze_time
from unittest.mock import MagicMock, call

import pytest
import requests

from nephthys.clients.requests import (
    catch_logger_exception,
    decorate_log_request,
    decorate_log_response,
    Session)
from nephthys.filters.requests import BODY_NOT_LOGGABLE


@pytest.fixture
def m():
    with requests_mock.Mocker() as m:
        yield m


def test_exception_catcher_decorator(caplog):
    caplog.set_level(logging.DEBUG, logger="requests_in")

    def on_my_funct():
        raise Exception()

    catch_logger_exception(on_my_funct)()

    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelno == logging.ERROR
    assert record.msg == "Failed to log"
    assert record.exc_info is not None


def test_decorate_log_request_headers():
    log_record = MagicMock()
    test_headers = {"key1": "value1", "key2": "value2"}
    request = MagicMock(url="https://ovalmoney.com", headers=test_headers)
    decorate_log_request(log_record, request)
    for name, value in test_headers.items():
        expected = call(name, value)
        assert expected in log_record.add_request_header.call_args_list


def test_decorate_log_request_querystring():
    log_record = MagicMock()
    name, value = "query", "var"
    request = MagicMock(url=f"https://ovalmoney.com?{name}={value}")
    request.headers = None
    decorate_log_request(log_record, request)
    expected = call(name, [value])
    assert expected in log_record.add_request_querystring.call_args_list


def test_decorate_log_request_body():
    log_record = MagicMock()
    request = MagicMock(url=f"https://ovalmoney.com")
    body = MagicMock()
    body.decode.return_value = "return"
    request.body = body
    decorate_log_request(log_record, request)
    assert log_record.request_body == "return"
    body.decode.assert_called_with("UTF-8", errors="strict")


def test_decorate_log_request_body_raises():
    log_record = MagicMock()
    body = MagicMock()
    body.decode.side_effect = UnicodeDecodeError("bla", bytes(0), 0, 1, "bla")
    request = MagicMock(url=f"https://ovalmoney.com", body=body)
    decorate_log_request(log_record, request)
    assert log_record.request_body == "<RAW Data>"
    body.decode.assert_called_with("UTF-8", errors="strict")


def test_decorate_log_response_headers():
    log_record = MagicMock()
    test_headers = {"key1": "value1", "key2": "value2"}
    response = MagicMock(url="https://ovalmoney.com", headers=test_headers)
    decorate_log_response(log_record, response)
    for name, value in test_headers.items():
        expected = call(name, value)
        assert expected in log_record.add_response_header.call_args_list


@pytest.mark.parametrize(
    "encoding,expected_encoding", [(None, "utf-8"), ("utf-8", "utf-8")]
)
def test_decorate_log_response_content(encoding, expected_encoding):
    log_record = MagicMock()
    content = MagicMock()
    content.decode.return_value = "test"
    response = MagicMock(
        url="https://ovalmoney.com", headers=None, encoding=encoding, content=content
    )
    decorate_log_response(log_record, response)
    content.decode.assert_called_with(expected_encoding, errors="strict")
    assert log_record.response_body == "test"


@pytest.mark.parametrize(
    "exception,encoding,expected_encoding",
    [
        (TypeError, None, "utf-8"),
        (UnicodeDecodeError, None, "utf-8"),
        (LookupError, "UtF8", "UtF8"),
    ],
)
def test_decorate_log_response_content_raises(exception, encoding, expected_encoding):
    log_record = MagicMock()
    content = MagicMock()
    content.decode.side_effect = exception
    response = MagicMock(
        url="https://ovalmoney.com", headers=None, encoding=encoding, content=content
    )
    decorate_log_response(log_record, response)
    content.decode.assert_called_with(expected_encoding, errors="strict")
    assert log_record.response_body == "<RAW Data>"


def test_send_log_record(caplog, m):
    caplog.set_level(logging.INFO)
    m.get("https://ovalmoney.com/user", status_code=200)

    s = Session()
    s._logger.info = MagicMock()
    s.get("https://ovalmoney.com/user")

    assert s._logger.info.called


def test_send_log_record_exception(caplog, m):
    caplog.set_level(logging.INFO)
    m.get("https://ovalmoney.com/user", exc=requests.exceptions.ConnectTimeout)

    s = Session()
    s._logger.exception = MagicMock()
    with pytest.raises(requests.exceptions.ConnectTimeout):
        s.get("https://ovalmoney.com/user")

    assert s._logger.exception.called


def test_raw_data_reponse_body_log(caplog, m):
    caplog.set_level(logging.INFO)
    m.get(
        "https://ovalmoney.com/raw_data",
        headers={"Content-Type": "video"},
        content=b"\xFF\x8F",
    )

    s = Session()
    s.get("https://ovalmoney.com/raw_data")

    log_rec = [rec for rec in caplog.records][0]

    assert log_rec.msg["response"]["body"] == BODY_NOT_LOGGABLE.format(
        log_rec.msg["response"]["header"]["Content-Type"]
    )


def test_raw_data_request_body_log(caplog, m):
    caplog.set_level(logging.INFO)
    m.post("https://ovalmoney.com/raw_data")

    s = Session()
    s.post("https://ovalmoney.com/raw_data", headers={"Content-Type": "video"}, data=b"\xFF\x8F")

    log_rec = [rec for rec in caplog.records][0]

    assert log_rec.msg["request"]["body"] == BODY_NOT_LOGGABLE.format(
        log_rec.msg["request"]["header"]["Content-Type"]
    )


def test_text_reponse_body_log(caplog, m):
    caplog.set_level(logging.INFO)
    m.get(
        "https://ovalmoney.com/text_data",
        headers={"Content-Type": "text/plain"},
        text="Response",
    )

    s = Session()
    s.get("https://ovalmoney.com/text_data")

    log_rec = [rec for rec in caplog.records][0]

    assert log_rec.msg["response"]["body"] == "Response"


def test_form_data_request_body_log(caplog, m):
    caplog.set_level(logging.INFO)
    m.post("https://ovalmoney.com/form_data")

    s = Session()
    s.post("https://ovalmoney.com/form_data", data={"key": "value", "key2": "value2"})

    log_rec = [rec for rec in caplog.records][0]

    assert log_rec.msg["request"]["body"] == "key=value&key2=value2"


def test_session_fail_http(m):
    m.get("https://ovalmoney.com/user", exc=requests.exceptions.ConnectTimeout)

    s = Session()
    with pytest.raises(requests.exceptions.ConnectTimeout):
        s.get("https://ovalmoney.com/user")


def test_session_fail_http_log(caplog, m):
    caplog.set_level(logging.INFO)
    m.get("https://ovalmoney.com/user", exc=requests.exceptions.ConnectTimeout)

    s = Session()
    with pytest.raises(requests.exceptions.ConnectTimeout):
        s.get("https://ovalmoney.com/user")

    log_rec = [rec for rec in caplog.records][0]

    assert isinstance(log_rec.exc_info[1], requests.exceptions.ConnectTimeout)
    log = log_rec.msg
    assert log["request"]["method"] == "GET"
    assert log["request"]["url"] == "https://ovalmoney.com/user"
    assert log["request"]["path"] == "/user"
    assert log["request"]["host"] == "ovalmoney.com"


def test_session_success(caplog, m):
    caplog.set_level(logging.INFO)
    m.get(
        "https://ovalmoney.com/user?key1=value1&key2=value2&key2=value3",
        json={"test": True},
        headers={"Content-Type": "application/json"},
        status_code=200,
    )

    s = Session()
    response = s.get("https://ovalmoney.com/user", params={"key1": "value1", "key2": ["value2", "value3"]})

    assert response.headers
    assert response.status_code == 200
    assert response.json() == {"test": True}


def test_session_success_log(caplog, m):
    caplog.set_level(logging.INFO)
    m.get(
        "https://ovalmoney.com/user?key1=value1&key2=value2&key2=value3",
        json={"test": True},
        headers={"Content-Type": "application/json"},
        status_code=200,
    )

    with freeze_time(datetime.utcnow()):
        s = Session(log_tag="test")
        s.get("https://ovalmoney.com/user", params={"key1": "value1", "key2": ["value2", "value3"]})
        now = datetime.utcnow().timestamp()

    log = [rec.msg for rec in caplog.records][0]

    assert log["extra_tags"] == ["test", "requests_out"]

    assert log["request"]["start"] == now
    assert log["request"]["end"] == now
    assert log["request"]["time"] == 0.0

    assert log["request"]["method"] == "GET"
    assert (
            log["request"]["url"]
            == "https://ovalmoney.com/user?key1=value1&key2=value2&key2=value3"
    )
    assert log["request"]["path"] == "/user"
    assert log["request"]["host"] == "ovalmoney.com"
    assert log["request"]["query"] == {"key1": "value1", "key2": "value2,value3"}

    assert log["response"]["body"] == '{"test": true}'
    assert log["response"]["status_code"] == 200



import logging
import copy
from .json import JSONRenderer


class PrettyFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        sort_keys = kwargs.pop("json_sort_keys", True)
        indent = kwargs.pop("json_indent", 2)
        self._renderer = JSONRenderer(indent=indent, sort_keys=sort_keys)

        super().__init__(*args, **kwargs)

    def format(self, record):

        if isinstance(record.msg, dict):
            record.message = self._renderer(copy.deepcopy(record.msg))
        else:
            record.message = record.getMessage()

        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)

        s = self.formatMessage(record)
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self.formatStack(record.stack_info)
        return s
